#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#Copyright (c) 2020-2024 - University of Liège - Digital Energy and Agriculture Lab (DEAL)
#Author : Nicolas De Cock (nicolas.decock1@gmail.com)
#This file is part of the PASE software, and is distributed under the MIT license.


import numpy as np
from vtk import vtkPointSet, vtkPoissonDiskSampler, vtkMaskPoints
import pyvista as pyV


"""
Class Mesh


This class aims to provide point sources which are used by the light_scene class 
in order to compute the diffuse and direct light at different points of the scene.
The class contains different "Add_" method which are specific specific for each application.
The user has to configure these methods and provide a flag which will be used
to latter on gather the points according to their nature. 


Example of use:
    
Initialization:
M = Mesh()

Addition of a regular mesh at the ground level:
M.add_plane_ground_regular_meshes(0,1,0,3,0.1,1)

Addition of a mesh from a PV panel geometry:
M.add_PV_mesh(first_panel)

Addition of a sensor:
M.add_sensor(0,1,1,"Pyranometre")

Retrieve of point sources generated by the previous "Add" methods:
S = M.get_sourcepoints()

Retrieve of dictionnary providing the mapping between the flagID and the flag
Dict = M.get_sources_flag_dict()


"""

class Mesh:
    def __init__(self,automerge = True):
        
        #Attribut par encore utilise mais ayant pour but de fusionner les sources ayant le meme flag en leur attribuant le meme flagID
        self.automerge = automerge
        self.sourcepoints = np.empty((0, 4))
        self.sources_flag_dict = {}
        self.flag_key = 0


# =============================================================================
#     PUBLIC METHODS
# =============================================================================
    


    def add_plane_ground_regular_meshes(self, X_min, X_max, Y_min, Y_max, X_increment, Y_increment,flag = "ground", zcoord = 0):
        """
        Public method creating a regular rectangular mesh which is added in the sourcepoints class attribut
        
        Parameters:
            X/Y_min,X/Y_max (float): Minimal and maximal distance of the X/Y coordinates
            X/Y_increment (float): Distance between each X/Y points
            flag  (str): name of the set of points
            zcoord (float): coordinate on the z axis of the plane

        Returns:
           
        """
        xrng = np.arange(X_min, X_max, X_increment)
        yrng = np.arange(Y_min, Y_max, Y_increment)
        zrng = zcoord
        # Create coordinate arrays for each dimension using meshgrid
        x_coords, y_coords, z_coords = np.meshgrid(xrng, yrng, zrng, indexing='ij')
        
        # Stack the coordinate arrays to create an array of coordinates
        coordinates = np.stack((x_coords, y_coords, z_coords), axis=-1)
        
        # Reshape the coordinates to a 2D array (number of coordinates, number of dimensions)
        coordinates_2d = coordinates.reshape(-1, coordinates.shape[-1])

        self.add_sourcepoints(coordinates_2d,flag)
   
    


    def add_PV_mesh(self, Geometry, flag="PV"):
        """
        Public method creating a mesh from a Photovoltaic panel geometry
        
        Parameters:
            Geometry (float): 2D pyvista.PolyData object of the photovoltaic installation
            radius (float): minimum distance allow between two samples points
            flag  (str): name of the set of points

        Returns:
           
        """
        
        #radius = 0.05
        #coordinates_2d = self.Get_Surface_Sample_Points(Geometry,radius)
        coordinates = self.get_PV_mesh(Geometry)
        self.add_sourcepoints(coordinates,flag)
        
        
    def add_sensor(self,x,y,z,flag="sensor"):
        """
        Public method creating a sensor located at the x,y,z coordinates
        
        Parameters:
            x,y,z (float): coordinates of the sensor
            flag  (str): name of the set of points
            
        Returns:
           
        """
        self.add_sourcepoints(np.array([[x,y,z]]),flag)




    def get_sourcepoints(self,GetFlagId=True,FlagId=[]):
        """
        Public methods providing the sourcepoints from all the mesh which have been added to the class instance
        
        Parameters:
            GetFlagId (bool): Return the last columns containing the flag ID (True/False)
            FlagId (list of int): if not empty, return only the sourcepoints within the FlagId list
        
        Returns:
            sourcepoints (np.ndarray): (n x 3) or (n x 4) ndarray containing the requested sourcepoints    
        """
        #Filter of the points based on a FlagId
        if len(FlagId)>0:
            Temp = self.sourcepoints[np.isin(self.sourcepoints[:,3], list(FlagId))]
        else:
            Temp = self.sourcepoints
            
        if GetFlagId:
            return Temp
        else:
            return Temp[:,:-1]
    
    

    def get_flagID_by_flag(self,Flags):
        """
        Public methods providing the FlagId related to a list of Flags name
        
        Parameters:
            Flags  (list/str): list of Flags name from which the point source are needed
       
        Returns:
            FlagId (list of int): FlagId link to the requested Flags 
        """
        if type(Flags) == str:
            l = []
            l.append(Flags)
        else:
            l = Flags
        return [key for key in self.sources_flag_dict if self.sources_flag_dict[key] in l]

    
  
    def get_sourcepoints_by_flag(self,Flags):
        """
        Public methods providing the sourcepoints  related to a list of Flags name
        
        Parameters:
            Flags  (list/str): list of Flags name from which the point source are needed
       
        Returns:
            sourcepoints (np.ndarray): (n x 4) ndarray containing the requested sourcepoints    
        """
        return self.get_sourcepoints(self,FlagId=self.get_flagID_by_flag(Flags))
        
    
 
    #Public
    def get_sources_flag_dict(self):
        """
        Public methods providing SourceDict for all the sources
        
        Parameters:
              
        Returns:
            SourceDict (dict): dictionnary providing the mapping between the FlagId (key) and the Flags (value)    
        """   
        return self.sources_flag_dict
    
    
# =============================================================================
#     PRIVATE METHODS
# =============================================================================
    
    def get_source_points_index(self,Flags):
        """
        Private methods providing the pointsource index based on a list of flags
        
        Parameters:
            Flags (list of str): Flag name(s) from which the index want to be retrieved
        Returns:
            Index (list of int): Index of the point sources for which the flags are in the input Flags  
        """   
  
        FlagId = self.get_flagID_by_flag(Flags)
        np.isin(self.sourcepoints[:,3], list(FlagId))
        return np.isin(self.sourcepoints[:,3], list(FlagId)).nonzero()[0].astype(int)
    
    def add_sourcepoints(self,coordinates,flag):
        
        """
        Private method appending pointsources in the pointsources class attribut
        
        Parameters:
            coordinates (np.ndarray (n x 3): x,y,z coordinates of the points source
            flag (str) : name of the set of points
        
        Returns:
        """    
        #Addition of the flagID column (fourth column)        
        coordinatesT = np.hstack((coordinates,np.ones((coordinates.shape[0], 1))*self.flag_key)) 
        self.sourcepoints = np.concatenate((self.sourcepoints, coordinatesT))
        self.sources_flag_dict[self.flag_key] = flag
        self.flag_key += 1
        
    
  
    def get_surface_sample_points(self,OPolyData,radius=0.05,Filter="Poisson",MaxSubdivide=4):
        """
        Private method used to retrieve source point from PolyData object
        
        Parameters:
            OPolyData (pyvista.PolyData): PolyData of the object from which the mesh will be done
            radius (float): minimum distance allow between two samples points
            Filter (str): type of filter used to get the sample points
            MaxSubdivide (int): maximum number of subdivision of the original mesh 
        
        Returns:
            x,y,z coordinates (np.ndarray)
        """  
        # PolyDataArea = abs((OPolyData.GetBounds()[0]-OPolyData.GetBounds()[1])*(OPolyData.GetBounds()[3]-OPolyData.GetBounds()[4]))
        # NCells = len(OPolyData.cell)
        SuperSampleData = OPolyData.subdivide(MaxSubdivide)
        if Filter == "Random":
            Mask = vtkMaskPoints()
            Mask.SetRandomModeType(4) #4 --> uniform sampling
            Mask.SetInputData(SuperSampleData)
            Mask.Update()
        elif Filter == "Poisson":
            Mask = vtkPoissonDiskSampler() #Algorithm using darts like sampling
            Mask.SetInputData(SuperSampleData)
            Mask.SetRadius(radius)
            Mask.Update()
        else:
            print('Error Filter not correctly set Poisson or Random')
        return pyV.PointSet(vtkPointSet().SafeDownCast(Mask.GetOutputDataObject(0))).points
    
    def get_PV_mesh(self, Polydata):
        corners = Polydata.points
        print(corners)
        vertex = [0,1,3,2]
        print(vertex)
        step = 0.1
        distA = np.linalg.norm(corners[vertex[0]] - corners[vertex[1]]) 
        distB = np.linalg.norm(corners[vertex[1]] - corners[vertex[2]]) 
        
        dirA = (corners[vertex[0]] - corners[vertex[1]]) / distA
        dirB = (corners[vertex[1]] - corners[vertex[2]]) / distB
        
        spaceA = np.linspace(0,distA,int(distA/step)+1)
        spaceB = np.linspace(0,distB,int(distB/step)+1)
        
        A = dirA.reshape(1,3)*spaceA.reshape(len(spaceA),1)
        B = dirB.reshape(1,3)*spaceB.reshape(len(spaceB),1)
        C = np.array([A+b for b in B]).reshape(len(A)*len(B),3)
        POINTS = corners[vertex[0]]-C
        return POINTS
